@@ -11,7 +11,7 @@ require Exporter;
 use Parse::RecDescent ();
 use YAML ();
 
-$VERSION = '0.01_07';
+$VERSION = '0.02_01';
 $VERSION = eval $VERSION;  # see L<perlmodstyle>
 
 # Global flags 
@@ -111,6 +111,15 @@ enum      :
     $return->{name} = $item[2] if $item[2];
     $return->{value} = $item[3] if $item[3];
     $return->{variable} = $item[4] if $item[4];
+    unless($item[3]) { 
+        # no enum body, possibly inside typedef or forward decl
+        # split variables from enum_name
+        my @v = split /(?<!,)\s+(?!,)/, $item[2];
+        if (@v > 1) {
+            $return->{variable} = pop @v;
+            $return->{name}     = join(" ", @v);
+        }
+    }
   }  
   { print STDERR $item[1], ": ", join(" ", $item[2], join(" ", @{$item[3]}), $item[4]), "\n" 
         if $::RD_DEBUG }
@@ -158,7 +167,7 @@ extern   :
     { print STDERR "extern: ", 
           join(" ", $return->{type}, $return->{subtype}), "\n" if $::RD_DEBUG } 
 namespace: 
-  keyword_namespace namespace_name '{' namespace_body(s?) '}'
+  keyword_namespace namespace_name '{' namespace_body(s?) '}' ( ';' | )
   { $return = { type => 'namespace', name => $item[2], body => [] } }
   { foreach my $a (@{$item[4]}) { push @{$return->{body}}, @$a }    }
   { print STDERR "namespace: ", $item[2], "\n" if $::RD_DEBUG }
@@ -170,20 +179,26 @@ class    :
     $return->{inheritance} = $item[3] if $item[3];
     $return->{body} = $item[4] if $item[4];
     $return->{variable} = $item[5] if $item[5];
+    unless($item[4]) { 
+        # no class body, possibly inside typedef or forward decl
+        # split variables from enum_name
+        my @v = split /(?<!,)\s+(?!,)/, $item[2];
+        if (@v > 1) {
+            $return->{variable} = pop @v;
+            $return->{name}     = join(" ", @v);
+        }
+    }
   } 
   { print STDERR $item[1][0], ": ", 
-        join(" ", @item[2 .. $#item-1]), "\n" if $::RD_DEBUG }
+        join(" ", @item[2 .. $#item-3]), "\n" if $::RD_DEBUG }
 # a simple trap here 
 # to prevent template function parsed as normal one
 function : 
     keyword_template <commit> <reject>
   | class_accessibility <commit> <reject> 
   | function_header function_body
-    { 
-      $return = { type => 'function', name => $item[1][1] };
-      $return->{fullname} = $item[1];
-    } 
-    { print STDERR "function: ", $item[1][1], "\n" if $::RD_DEBUG }
+    { $return = { type => 'function', %{$item[1]} }; } 
+    { print STDERR "function: ", $item[1]->{name}, "\n" if $::RD_DEBUG }
 # QT-specific macros
 qt_macro_1 : 
   'QT_BEGIN_HEADER' | 'QT_END_HEADER' | 'Q_OBJECT' | 'Q_GADGET' 
@@ -329,35 +344,221 @@ variables : next_semicolon { $return = $item[1] } | { $return = '' }
 # function related
 # at least one '()' block should appear for a valid header
 # trap other keywords to prevent mess
-# return a structure [ attribute_prefix, function_name,
-# .., function_macro ]
-# the idea here is to make sure the second one is _ALWAYS_ function name
 function_header       : 
     (   keyword_comment | keyword_class | keyword_enum 
       | keyword_typedef ) <commit> <reject>
   | function_header_block(s) 
-    { $return = $item[1][0] !~ m/^\_\_/io ? [ '', @{$item[1]} ] : $item[1] } 
-    { if ($return->[1] =~ m/operator\(\)$/o) { 
-          $return->[1] .= splice @$return, 2, 1; }                         }
+    { 
+      $return->{name} = ''; 
+      foreach my $i (@{$item[1]}) { 
+          if ($i->{_subtype} == 1) { 
+              # attribute
+              $return->{name} .= $i->{_value}; 
+          } elsif ($i->{_subtype} == 2) { 
+              # function name with params
+              $return->{name} .= $i->{_name};
+              $return->{parameter} = $i->{_value} if 
+                  @{$i->{_value}};
+          } elsif ($i->{_subtype} == 3) { 
+              # other macros
+              push @{$return->{property}}, $i->{_value} if 
+                  $i->{_value};
+          } 
+      } 
+    } 
 function_header_next_token : 
   next_bracket_or_brace_or_semicolon 
     { $return = $item[1] } 
   | { $return = ''       } 
-function_macro_99     : 
-    'const' 
-  | qt_macro_99 
-  | kde_macro_99 
+# old backup 
+#function_header_block_old_unused : 
+#  function_header_next_token '(' function_header_loop(s?) ')' 
+#    { $return = join("", $item[1], $item[2], @{$item[3]}, $item[4]) }
+#  | function_macro_99(s?) 
+#    { $return = join("", @{$item[1]}) } 
+
+# function parameter process
+# parse parameters and simply consume __attribute__ 
 function_header_block : 
-  function_header_next_token '(' function_header_loop(s?) ')' 
-    { $return = join("", $item[1], $item[2], @{$item[3]}, $item[4]) }
+  (   function_header_next_token 
+      { $item[1] =~ m/\_\_attribute\_\_\s*$/o ? 1 : undef } 
+      '(' function_header_loop(s?) ')' 
+      { 
+        $return = { _subtype => 1, 
+                    _value   => join("", $item[1], $item[3], 
+                                    @{$item[4]}, $item[5]) }; 
+      } 
+    | function_header_next_token 
+      { $item[1] =~ m/^\:/o ? 1 : undef } 
+      function_header_loop(s) 
+      { $return = { _subtype => 0 }; } 
+    | function_header_next_token 
+      { $item[1] =~ m/operator$/o ? 1 : undef } 
+      '(' ')' '(' ')' 
+      { 
+        $return = { _subtype => 2, _name => $item[1].'()', }; 
+        $return->{_value} = []; 
+      } 
+    | function_header_next_token 
+      '(' 
+      ( function_parameters { $return = $item[1]; } | { $return = []; } ) 
+      ')' 
+      { 
+        $return = { _subtype => 2, _name => $item[1], };  
+        $return->{_value} = $item[3]; 
+      } 
+  ) { $return = $item[1]; } 
   | function_macro_99(s?) 
-    { $return = join("", @{$item[1]}) } 
+    { $return = { _subtype => 3, _value => join("", @{$item[1]}) }; } 
+# TODO: loop in more elegant way
 function_header_loop  : 
-    function_header_next_token ( '(' function_header_loop(s) ')' 
+  (   function_header_next_token { $return = $item[1]; } 
+    | { $return = ''; } 
+  ) 
+  (   '(' ')' { $return = '()'; } 
+    | '(' function_header_loop(s?) ')' 
       { $return = join("", $item[1], @{$item[2]}, $item[3]) } 
-    | { $return = '' } ) 
-    { $return = join("", @item[1 .. $#item]) } 
-  | { $return = '' } 
+    | { $return = '' } 
+  ) 
+  { $return = join("", @item[1 .. $#item]) } 
+
+function_macro_99     : 
+    'const' | qt_macro_99 | kde_macro_99 
+
+function_parameters : 
+    function_parameter_loop 
+
+function_parameter_loop : 
+    function_parameter 
+    (   ',' function_parameter_loop { $return = $item[2] } 
+      | { $return = [] } 
+    ) 
+    { $return = [ $item[1] ]; push @$return, @{$item[2]} if @{$item[2]} } 
+
+function_parameter  : 
+    function_parameter_declaration function_parameter_default_value(?) 
+    { 
+      $return = $item[1]; 
+      $return->{default} = $item[2]->[0] if @{$item[2]}; 
+    } 
+
+function_parameter_declaration_next_token : 
+    /(?>[^\,\=\(\)\<\>]+)/iso { ( $return = $item[1] ) =~ s/\n/ /go } 
+
+function_parameter_declaration            : 
+    (   function_parameter_declaration_next_token 
+        { $return = $item[1]; } 
+      | { $return = '';       } 
+    ) 
+    (   function_parameter_function_pointer 
+        { $return = { subtype => 'fpointer', value => $item[1] }; } 
+      | function_parameter_template_type 
+        { $return = { subtype => 'template', value => $item[1] }; } 
+      | { $return = { subtype => 'simple' }; } 
+    ) 
+    { 
+      if ($item[2]->{subtype} eq 'simple') {
+          $return = { name => $item[1], subtype => 'simple', };
+      } elsif ($item[2]->{subtype} eq 'template') { 
+          $return = { name => $item[1]. $item[2]->{value} };
+          $return->{subtype} = 'template'; 
+      } elsif ($item[2]->{subtype} eq 'fpointer') { 
+          $return = $item[2]->{value};
+          $return->{name} = $item[1];
+          $return->{subtype} = 'fpointer'; 
+      } 
+    } 
+
+function_parameter_template_type_next_token : 
+    /(?>[^\<\>]+)/iso { ( $return = $item[1] ) =~ s/\n/ /go } 
+
+function_parameter_template_type            : 
+    '<' function_parameter_template_type_loop(s) '>' 
+    (   function_parameter_declaration_next_token 
+        { $return = $item[1]; } 
+      | { $return = '';       } 
+    ) 
+    { 
+      $return = join(" ", $item[1], 
+          join("", @{$item[2]}), $item[3], $item[4]); 
+    } 
+
+function_parameter_template_type_loop       : 
+    (   function_parameter_template_type_next_token 
+        { $return = $item[1]; } 
+      | { $return = '';       } 
+    ) 
+    (   '<' function_parameter_template_type_loop '>' 
+        { $return = join(" ", @item[1 .. 3]); } 
+      | { $return = '';                      } 
+    ) 
+    { $return = join("", @item[1 .. 2]); } 
+
+function_parameter_function_pointer_next_token : 
+    /(?>[^\(\)]+)/iso { ( $return = $item[1] ) =~ s/\n/ /go } 
+
+function_parameter_function_pointer            : 
+    '(' function_parameter_function_pointer_loop ')' 
+    '(' function_parameter_loop ')' 
+    { $return = { name => $item[2], parameter => $item[5], }; }
+
+function_parameter_function_pointer_loop       : 
+    (   function_parameter_function_pointer_next_token 
+        { $return = $item[1]; }
+      | { $return = '';       } 
+    ) 
+    (   function_parameter_function_pointer 
+        { $return = $item[1]; } 
+      | { $return = {};       } 
+    ) 
+    { 
+      $return = { name => $item[1], }; 
+      if (exists $item[2]->{name}) { 
+          $return->{name}  = $item[2]->{name}; 
+          $return->{parameter} = $item[2]->{parameter};
+      }
+    } 
+
+function_parameter_default_value_next_token : 
+    /(?>[^\(\'\"\)\,]+)/iso { ( $return = $item[1] ) =~ s/\n/ /go } 
+
+function_parameter_default_value            : 
+    '=' function_parameter_default_value_loop(s) 
+    { $return = join("", @{$item[2]}); } 
+    { print STDERR "default value:", $return, "\n" if $::RD_DEBUG; } 
+
+function_parameter_default_value_loop_token_dispatch : 
+    '(' ')' 
+    { $return = '()'; } 
+  | '(' function_parameter_default_value_loop2 ')' 
+    { $return = join(" ", @item[1 .. 3]); } 
+  | "'" /(?>[^\']*)/iso "'" { $return = join("", @item[1 .. 3]); } 
+  | '"' /(?>[^\"]*)/iso '"' { $return = join("", @item[1 .. 3]); } 
+
+function_parameter_default_value_loop2      : 
+    (   function_parameter_default_value_next_token 
+        { $return = $item[1]; } 
+      | { $return = '';       } 
+    ) 
+    (   function_parameter_default_value_loop_token_dispatch 
+        { $return = $item[1]; } 
+      | ',' function_parameter_default_value_loop2 
+        { $return = join(" ", @item[1 .. 2]); } 
+      | { $return = '';                       } 
+    ) 
+    { $return = join("", @item[1 .. 2]); } 
+
+function_parameter_default_value_loop       : 
+    (   function_parameter_default_value_next_token 
+        { $return = $item[1]; } 
+      | { $return = '';       } 
+    ) 
+    (   function_parameter_default_value_loop_token_dispatch 
+        { $return = $item[1]; } 
+      | { $return = '';       } 
+    ) 
+    { $return = join('', @item[1 .. 2]); } 
+
 function_body         : 
     ';' { $return = '' } 
   | '=' '0' ';' { $return = '' }
@@ -371,7 +572,7 @@ balanced_brace            :
 
 # enum related
 enum_name          : 
-    next_begin_brace { $return = $item[1] }
+    next_brace_or_semicolon { $return = $item[1] }
   | { $return = ''       } 
 # enum_unit(s /,/) _NOT_ work here
 enum_body          : 
@@ -379,6 +580,7 @@ enum_body          :
   | '{' enum_unit(s) '}'
     { $return = $item[2] }
     #{ print STDERR "enum_body: ", join(" ", @{$item[2]}), "\n" if $::RD_DEBUG } 
+  | { $return = ''       }
 enum_unit          : 
   next_dot_or_end_brace ( ',' | )
   { $return = (split /=/, $item[1])[0] }
